#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai
from __future__ import (unicode_literals, division, absolute_import,
                        print_function)

__license__   = 'GPL v3'
__copyright__ = '2011, Pr.BarnArt based on the Barnes work by Grant Drake'
__docformat__ = 'restructuredtext en'

import time
from urllib import quote
from Queue import Queue, Empty

from lxml.html import fromstring, tostring

from calibre import as_unicode
from calibre.ebooks.metadata import check_isbn
from calibre.ebooks.metadata.sources.base import Source
from calibre.utils.icu import lower
from calibre.utils.cleantext import clean_ascii_chars
from calibre.utils.localization import get_udc
#from calibre import ipython

class catawiki(Source):

    name                    = 'catawiki'
    description             = _('Downloads metadata en covers van catawiki.nl')
    author                  = 'DrMerry Based on Pr. BarnArt'
    version                 = (0, 0, 2)
    minimum_calibre_version = (0, 8, 0)

    capabilities = frozenset(['identify', 'cover'])
    touched_fields = frozenset(['title', 'authors', 'identifier:isbn', 'rating', 'comments', 'publisher', 'pubdate'])
    has_html_comments = True
    supports_gzip_transfer_encoding = True
    cached_cover_url_is_reliable = True
    '''
    Zoeken naar heer bommel
    http://www.catawiki.nl/search?type=&q=Heer+bommel&commit=Zoeken

    http://www.bol.com/nl/s/algemeen/zoekresultaten/Ntt/heer+bommel/Ntk/media_all/Nty/1/N/0/Ne/0/search/true/searchType/qck/index.html?_requestid=162858

    '''
    BASE_URL0 = 'http://www.catawiki.nl'
    BASE_URL = 'http://www.catawiki.nl/search?type=&q='
    '''http://www.bol.com/nl/s/boeken/zoekresultaten/Ntt/'''
    BASE_ISBN= '&commit=Zoeken'
    #/Nty/1/search/true/searchType/adv/section/books/N/8293/Ntk/isbncode/index.html?_requestid=116710' 
    BASE_URL_1 = ''
    '''/Ntk/nl_books_all/Nty/1/N/8299+8293/Ne/8299+8293'''
    BASE_URL_LAST = '&commit=Zoeken'
    '''/search/true/searchType/qck/index.html?_requestid=17572' '''
    
    def config_widget(self):
        '''
        Overriding the default configuration screen for our own custom configuration
        '''
        from calibre_plugins.catawiki.config import ConfigWidget
        return ConfigWidget(self)

    def get_book_url(self, identifiers):
        catawiki_nl_id = identifiers.get('catawiki_nl', None)
        if catawiki_nl_id:
            return ('catawiki_nl', catawiki_nl_id,
                  '%s%s%s'%(catawiki_nl_id))
                  

    def create_query(self, log, title=None, authors=None, identifiers={}, isbntobetested=True, testtitle=True, testauthor=True):

        isbn = check_isbn(identifiers.get('isbn', None))        
        q = ''
        if isbn is not None and isbntobetested:            
            return '%s%s%s'%(catawiki.BASE_URL,isbn,catawiki.BASE_ISBN)
        if title and testtitle:           
            # catawiki  maybe doesn't cope very well with non ascii names so convert
            title = get_udc().decode(title)
            title_tokens = list(self.get_title_tokens(title,
                                strip_joiners=False, strip_subtitle=True))
            if title_tokens:
                tokens = [quote(t.encode('utf-8') if isinstance(t, unicode) else t) for t in title_tokens]               
                q='+'.join(tokens)
                
            
        if authors and testauthor:
            # catawiki  maybe doesn't cope very well with non ascii names so convert
            authors = [get_udc().decode(a) for a in authors]
            author_tokens = self.get_author_tokens(authors,
                    only_first_author=True)
            if author_tokens:
                tokens = [quote(t.encode('utf-8') if isinstance(t, unicode) else t) for t in author_tokens]
                q+='+'+'+'.join(tokens)
            
        if not q:
            return None
        return '%s%s%s%s'%(catawiki.BASE_URL,q,catawiki.BASE_URL_1,catawiki.BASE_URL_LAST) 

    def get_cached_cover_url(self, identifiers):
        url = None
        catawiki_nl_id = identifiers.get('catawiki_nl', None)
        if catawiki_nl_id is None:
            isbn = identifiers.get('isbn', None)
            if isbn is not None:
               catawiki_nl_id = self.cached_isbn_to_identifier(isbn)
        if catawiki_nl_id is not None:
            url = self.cached_identifier_to_cover_url(catawiki_nl_id)
        return url

    def cached_identifier_to_cover_url(self, id_):
        with self.cache_lock:
            url = self._get_cached_identifier_to_cover_url(id_)
            if not url:
                # Try for a "small" image in the cache
                url = self._get_cached_identifier_to_cover_url('small/'+id_)
            return url

    def _get_cached_identifier_to_cover_url(self, id_):
        # This must only be called once we have the cache lock
        url = self._identifier_to_cover_url_cache.get(id_, None)
        if not url:
            # We could not get a url for this particular B&N id
            # However we might have one for a different isbn for this book
            # Barnes & Noble are not very consistent with their covers and
            # it could be that the particular ISBN we chose does not have
            # a large image but another ISBN we retrieved does.
            key_prefix = id_.rpartition('/')[0]
            for key in self._identifier_to_cover_url_cache.keys():
                if key.startswith('key_prefix'):
                    return self._identifier_to_cover_url_cache[key]
        return url

    def identify(self, log, result_queue, abort, title=None, authors=None,
            identifiers={}, timeout=30):
        '''
        Note this method will retry without identifiers automatically if no
        match is found with identifiers.
        '''
        matches = []
        #need to read the ratings of an indexpage, i cannot trace them on the productpage
        #use ratings for the selected matches, so the order of rating is equal to matches 
        ratings = []
        # If we have a catawiki id then we do not need to fire a "search"
        # at catawiki.nl. Instead we will go straight to the URL for that book.
        catawiki_nl_id = identifiers.get('catawiki_nl', None)
        isbn = check_isbn(identifiers.get('isbn', None))        
        br = self.browser

        if catawiki_nl_id:            
            matches.append('%s%s%s%s'%(catawiki_nl_id))
            #matches.append('%s%s%s%s'%(catawiki.BASE_URL, catawiki_nl_id,catawiki.BASE_URL_1,catawiki.BASE_URL_LAST))
        else:           
            query = self.create_query(log, title=title, authors=authors,
                    identifiers=identifiers, isbntobetested=True, testtitle=False, testauthor=False)            
            if query is None:
                log.error('Insufficient metadata to construct query')
                return
            isbn_match_failed = False
            log.info('Querying isbn: %s'%query)
            response = br.open_novisit(query, timeout=timeout)
            if isbn:
                try:
                    raw = response.read().strip()           
                    raw = raw.decode('utf-8', errors='replace')                    
                    if not raw:
                        log.error('Failed to get raw result for query: %r'%query)
                        return
                    root = fromstring(clean_ascii_chars(raw))
                    url_node=root.xpath('//div [@class="content"]/div/div/div [@class="cell"]/a /@href')
                    if url_node:
                        isbn_url=url_node[0]
                        rating='0'
                        #get rating
                        rtxt='//div [@class="title"]/span [@class="title"] /a  [@href="' + url_node[0] + '"]/../../p [@class="small_details"]/img /@title'
                        rating_node=root.xpath(rtxt)
                        if rating_node:
                            rating_title=rating_node[0]
                            rating=rating_title.rpartition(':')[2].strip()
                            end=rating.find(' ')
                            rating=rating[0:end]
                        
                        #end get rating
                        ratings.append(rating)
                        matches.append('%s%s'%(catawiki.BASE_URL0,isbn_url))
                        #http://www.bol.com/nl/p/nederlandse-boeken/de-derde-persoon/1001004006415949/index.html
                    else:
                        isbn_match_failed = True
                except:
                    msg = 'Failed to parse catawiki.nl page for query: %r'%query
                    log.exception(msg)
                    return msg        
            
            # For successful ISBN based searches we have already done everything we need to
            # So anything from this point below is for title/author based searches.
            if not isbn or isbn_match_failed:
                query = self.create_query(log, title=title, authors=authors,
                    identifiers=identifiers, isbntobetested=False, testtitle=True, testauthor=False)   
                log.info('Querying title: %s'%query)
                title_match_failed = False
                try:                    
                    raw = response.read().strip()
                    raw = raw.decode('utf-8', errors='replace')
                    if not raw:
                        log.error('Failed to get raw result for 2nd query: %r'%query)
                        title_match_failed = True
                        return
                    root = fromstring(clean_ascii_chars(raw))
                except:
                    msg = 'Failed to parse catawiki.nl page for query: %r'%query
                    log.exception(msg)
                    return msg
                # Now grab the matches from the search results, provided the
                # title and authors appear to be for the same book
                # self._parse_search_results(log, title, authors, root, matches,ratings,timeout)
                
                if title_match_failed:
                    query = self.create_query(log, title=title, authors=authors,
                        identifiers=identifiers, isbntobetested=False, testtitle=False, testauthor=True)            
                    log.info('Querying author: %s'%query)
                    try:                    
                        raw = response.read().strip()
                        raw = raw.decode('utf-8', errors='replace')
                        if not raw:
                            log.error('Failed to get raw result for 3rd query: %r'%query)
                            return
                        root = fromstring(clean_ascii_chars(raw))
                    except:
                        msg = 'Failed to parse catawiki.nl page for 3rd query: %r'%query
                        log.exception(msg)
                        return msg
                # Now grab the matches from the search results, provided the
                # title and authors appear to be for the same book
                self._parse_search_results(log, title, authors, root, matches,ratings,timeout)

        if abort.is_set():
            return
        if not matches:           
            if identifiers and title and authors:
                log.info('No matches found with identifiers, retrying using only'
                        ' title and authors')
                return self.identify(log, result_queue, abort, title=title,
                        authors=authors, timeout=timeout)
            log.error('No matches found with query: %r'%query)
            return

        
                
        from calibre_plugins.catawiki.worker import Worker
        #need to mix url and ratings for the worker class
        combos= zip(matches,ratings)
        workers = [Worker(combo, result_queue, br, log, i, self) for i, combo in
                enumerate(combos)]
        

        for w in workers:
            w.start()
            # Don't send all requests at the same time
            time.sleep(0.1)

        while not abort.is_set():
            a_worker_is_alive = False
            for w in workers:
                w.join(0.2)
                if abort.is_set():
                    break
                if w.is_alive():
                    a_worker_is_alive = True
            if not a_worker_is_alive:
                break

        return None

    def _parse_search_results(self, log, orig_title, orig_authors, root, matches, ratings, timeout):
        results = root.xpath('//a[@class="link_title"]')
        #results_url=root.xpath('//a class="link_title" [@href]')
        for result in results:
           result_url=result.xpath('@href')                       
        if not results:
            return        

        def ismatch(title):
             #only if the title exact matches
            match = False
            if lower(title)==lower(orig_title):
               match= True
            
            return match 
           
        import calibre_plugins.catawiki.config as cfg
        max_results = cfg.plugin_prefs[cfg.STORE_NAME][cfg.KEY_MAX_DOWNLOADS]
        for result in results:
            # Annoyingly, there isn't consistency in the search results
            # Sometimes the next level will be in <section> tags, sometimes not.
            title = result.text_content().replace('\n',' ').strip()
            while '  ' in title:
                title = title.replace('  ',' ')
            # Strip off any series information from the title
            if '(' in title:
                title = title.rpartition('(')[0].strip()
                title_tokens = list(self.get_title_tokens(orig_title))
            if not ismatch(title):
                log.error('Rejecting as not close enough match: %s '%(title))
                
            else:
                # Validate that the cat  is  not one we are not interested in    
                txt='//div [@class="product_title"]/a [@class="link_title"  and @href="' + result.xpath('@href')[0] + '"]/../../div [@class="subtitle"]'    
                cat_details = root.xpath(txt)
                valid_cat = True
                for c in cat_details:
                    cat = c.text_content().strip().lower()
                    if cat.find('gesproken woord') >= 0 : 
                        valid_cat = False
                    if cat.find(' cd ' ) >= 0:
                        valid_cat = False
                    if cat.find('luisterboek') >= 0:
                        valid_cat = False
                    if cat.find('voorgelezen') >= 0:
                        valid_cat = False
                    if valid_cat == False:
                        break
                rating='0'
                rtxt='//div [@class="product_title"]/a [@class="link_title"  and @href="' + result.xpath('@href')[0] + '"]/../../p [@class="small_details"]/img /@title'
                rating_node=root.xpath(rtxt)
                if rating_node:
                    rating_title=rating_node[0]
                    rating=rating_title.rpartition(':')[2].strip()
                    end=rating.find(' ')
                    rating=rating[0:end]
                                    
                if valid_cat:
                    result_url = catawiki.BASE_URL0 + result.xpath('@href')[0]                
                    matches.append(result_url)    
                    ratings.append(rating)
                    if len(matches) >= max_results:
                        break
        
            

    def download_cover(self, log, result_queue, abort,
            title=None, authors=None, identifiers={}, timeout=30):
        cached_url = self.get_cached_cover_url(identifiers)
        if cached_url is None:
            log.info('No cached cover found, running identify')
            rq = Queue()
            self.identify(log, rq, abort, title=title, authors=authors,
                    identifiers=identifiers)
            if abort.is_set():
                return
            results = []
            while True:
                try:
                    results.append(rq.get_nowait())
                except Empty:
                    break
            results.sort(key=self.identify_results_keygen(
                title=title, authors=authors, identifiers=identifiers))
            for mi in results:
                cached_url = self.get_cached_cover_url(mi.identifiers)
                if cached_url is not None:
                    break
        if cached_url is None:
            log.info('No cover found')
            return

        if abort.is_set():
            return
        br = self.browser
        log('Downloading cover from:', cached_url)
        try:
            cdata = br.open_novisit(cached_url, timeout=timeout).read()
            result_queue.put((self, cdata))
        except:
            log.exception('Failed to download cover from:', cached_url)


if __name__ == '__main__': # tests
    # To run these test use:
    # calibre-debug -e __init__.py
    from calibre.ebooks.metadata.sources.test import (test_identify_plugin,
            title_test, authors_test, series_test)
    test_identify_plugin(catawiki.name,
        [

            ( # A book with no ISBN specified
                {'title':"Harry Potter and the Sorcerer's Stone", 'authors':['J.K. Rowling']},
                [title_test("Harry Potter and the Sorcerer's Stone",
                    exact=True), authors_test(['J. K. Rowling']),
                    series_test('Harry Potter', 1.0)]

            ),

            ( # A book with an ISBN
                {'identifiers':{'isbn': '9780439064866'},
                    'title':'Chamber of Secrets', 'authors':['J.K. Rowling']},
                [title_test('Harry Potter and the Chamber of Secrets',
                    exact=True), authors_test(['J. K. Rowling']),
                    series_test('Harry Potter', 2.0)]

            ),

            ( # A book with a Barnes & Noble id
                {'identifiers':{'catawiki': '61-Hours/Lee-Child/e/9780440243694'},
                    'title':'61 Hours', 'authors':['Lee Child']},
                [title_test('61 Hours',
                    exact=True), authors_test(['Lee Child']),
                    series_test('Jack Reacher', 14.0)]

            ),

        ])


